from collections import namedtuple
from datagrep.policy import Policy
from datagrep.quality import Quality

SenseResult = namedtuple(
    "SenseResult",
    ["observation", "available_actions", "available_action_probabilities"],
)

ActResult = namedtuple("ActResult", ["action", "reward", "done"],)


def default_observation_wrapper(observation, available_actions):
    available_action_probabilities = [1.0 / len(available_actions)] * len(
        available_actions
    )
    return observation, available_actions, available_action_probabilities


class Agent:
    def __init__(
        self,
        observation_wrapper=default_observation_wrapper,
        reward_wrapper=lambda reward: reward,
        transition_function=None,
        policy_function=Policy(),
        quality_function=Quality(),
    ) -> None:
        self.observation_wrapper = observation_wrapper
        self.reward_wrapper = reward_wrapper
        self.transition_function = transition_function
        self.policy_function = policy_function
        self.quality_function = quality_function

    def get_policy_function(self):
        return self.policy_function

    def get_quality_function(self):
        return self.quality_function

    def sense(self, environment):
        (
            observation,
            available_actions,
            available_action_probabilities,
        ) = environment.observe(self.observation_wrapper)
        return SenseResult(
            observation, available_actions, available_action_probabilities
        )

    def act(
        self,
        environment,
        observation,
        available_actions,
        available_action_probabilities=None,
    ):
        action = self.policy_function.predict_action(
            observation, available_actions, available_action_probabilities
        )
        reward, done = environment.step(action, self.reward_wrapper)
        return ActResult(action, reward, done)

    def learn(self, trajectories, gamma=1.0, alpha=1.0):
        self.policy_function.learn(trajectories, gamma, alpha)
        self.quality_function.learn(trajectories, gamma, alpha)
